(**
 * Copyright (c) 2015, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the "hack" directory of this source tree.
 *
 *)

open Core_kernel
open Utils
open Typing_defs

module Reason = Typing_reason
module Unify = Typing_unify
module Env = Typing_env
module Subst = Decl_subst
module TUtils = Typing_utils
module SN = Naming_special_names
module Phase = Typing_phase
module TL = Typing_logic

type reactivity_extra_info = {
  method_info: ((* method_name *) string * (* is_static *) bool) option;
  class_ty: phase_ty option;
  parent_class_ty: phase_ty option
}

let empty_extra_info = {
  method_info = None;
  class_ty = None;
  parent_class_ty = None
}

module ConditionTypes = struct
  let try_get_class_for_condition_type (env: Env.env) (ty: decl ty) =
    match TUtils.try_unwrap_class_type ty with
    | None -> None
    | Some (_, ((_, x) as sid), _) ->
    begin match Env.get_class env x with
    | None -> None
    | Some cls -> Some (sid, cls)
    end

  let try_get_method_from_condition_type
    (env: Env.env)
    (ty: decl ty)
    (is_static: bool)
    (method_name: string) =
    match try_get_class_for_condition_type env ty with
    | Some (_, cls) ->
      let m = if is_static then cls.tc_smethods else cls.tc_methods in
      SMap.get method_name m
    | None -> None


  let localize_condition_type (env: Env.env) (ty: decl ty): locl ty =
    (* if condition type is generic - we cannot specify type argument in attribute.
       For cases when we check if containing type is a subtype of condition type
       if condition type is generic instantiate it with TAny's *)
    let ty =
      match try_get_class_for_condition_type env ty with
      | None -> ty
      | Some (_, cls) when cls.tc_tparams = [] -> ty
      | Some (((p, _) as sid), cls) ->
      let params =
        List.map cls.tc_tparams
          ~f:(fun (_, (p, x), _, _) -> Reason.Rwitness p, Tgeneric x) in
      let subst =
        Decl_instantiate.make_subst cls.tc_tparams [] in
      let ty = Reason.Rwitness p, (Tapply (sid, params)) in
      Decl_instantiate.instantiate subst ty in
    let ety_env = Phase.env_with_self env in
    let _, t = Phase.localize ~ety_env env ty in
    t
end

(* Given a pair of types `ty_sub` and `ty_super` attempt to apply simplifications
 * and add to the accumulated constraints in `constraints` any necessary and
 * sufficient [(t1,ck1,u1);...;(tn,ckn,un)] such that
 *   ty_sub <: ty_super iff t1 ck1 u1, ..., tn ckn un
 * where ck is `as` or `=`. Essentially we are making solution-preserving
 * simplifications to the subtype assertion, for now, also generating equalities
 * as well as subtype assertions, for backwards compatibility with use of
 * unification.
 *
 * If `constraints = []` is returned then the subtype assertion is valid.
 *
 * If the subtype assertion is unsatisfiable then return `failed = Some f`
 * where `f` is a `unit-> unit` function that records an error message.
 * (Sometimes we don't want to call this function e.g. when just checking if
 *  a subtype holds)
 *
 * If deep=true, elide singleton unions, treat invariant generics as both-ways
 * subtypes, and actually chase hierarchy for extends and implements.
 * For now, deep=false is used by subtype solving (conservative) and
 * deep=true is used for constraint decomposition (for instanceof).
 *
 * Annoyingly, we need to pass env back too, because Typing_phase.localize
 * expands type constants. (TODO: work out a better way of handling this)
 *
 * Special cases:
 *   If assertion is valid (e.g. string <: arraykey) then
 *     result can be the empty list (i.e. nothing is added to the result)
 *   If assertion is unsatisfiable (e.g. arraykey <: string) then
 *     we record this in the failed field of the result.
 *)
let with_error (f : unit -> unit) ((env, p) : (Env.env * TL.subtype_prop))
  : Env.env * TL.subtype_prop =
  env, TL.conj p (TL.Unsat f)

(* If `b` is false then fail with error function `f` *)
let check_with b f r = if not b then with_error f r else r

let valid env : Env.env * TL.subtype_prop = env, TL.valid

let (&&&) (env, p1) (f : Env.env -> Env.env * TL.subtype_prop) =
  if TL.is_unsat p1
  then env, p1
  else
    let env, p2 = f env in
    env, TL.conj p1 p2

let (|||) (env, p1) (f : Env.env -> Env.env * TL.subtype_prop) =
  if TL.is_valid p1
  then env, p1
  else
    let env, p2 = f env in
    env, TL.disj p1 p2

let if_unsat (f : unit -> Env.env * TL.subtype_prop) (env, p) =
  if TL.is_unsat p
  then f ()
  else env, p

let log_prop env prop =
  if TypecheckerOptions.log_inference_constraints Env.(env.genv.tcopt) then
    let p_as_string = Typing_print.subtype_prop env prop in
    let pos = Pos.string (Pos.to_absolute env.Env.pos) in
    let size = TL.size prop in
    let n_disj = TL.n_disj prop in
    let n_conj = TL.n_conj prop in
    TypingLogger.InferenceCnstr.log p_as_string ~pos ~size ~n_disj ~n_conj

(** Check that a mutability type is a subtype of another mutability type *)
let check_mutability
  ~(is_receiver: bool)
  (p_sub : Pos.t)
  (mut_sub: param_mutability option)
  (p_super : Pos.t)
  (mut_super: param_mutability option)
  env =
  let str m =
    match m with
    | None -> "immutable"
    | Some Param_owned_mutable -> "owned mutable"
    | Some Param_borrowed_mutable -> "mutable"
    | Some Param_maybe_mutable -> "maybe-mutable" in
  (* maybe-mutable <------immutable
                      |
                       <--mutable <-- owned mutable  *)
  match mut_sub, mut_super with
  (* immutable is not compatible with mutable *)
  | None, Some (Param_borrowed_mutable | Param_owned_mutable)
  (* mutable is not compatible with immutable  *)
  | Some (Param_borrowed_mutable | Param_owned_mutable), None
  (* borrowed mutable is not compatible with owned mutable *)
  | Some Param_borrowed_mutable, Some Param_owned_mutable
  (* maybe-mutable is not compatible with immutable/mutable *)
  | Some Param_maybe_mutable, (None | Some (Param_borrowed_mutable | Param_owned_mutable))
    ->
    env, TL.Unsat (fun () -> Errors.mutability_mismatch
      ~is_receiver p_sub (str mut_sub) p_super (str mut_super))
  | _ ->
    valid env

let empty_seen = Some SSet.empty

let log_subtype ~this_ty function_name env ty_sub ty_super =
  Typing_log.(log_with_level env "sub" 2 begin fun () ->
    let types =
      [Log_type ("ty_sub", ty_sub);
       Log_type ("ty_super", ty_super)]  in
    let types = Option.value_map this_ty ~default:types
      ~f:(fun ty -> Log_type ("this_ty", ty) :: types) in
    log_types (Reason.to_pos (fst ty_sub)) env
      [Log_head ("Typing_subtype." ^ function_name, types)]
  end)

(* Process the constraint proposition *)
let rec process_simplify_subtype_result ~this_ty ~fail env prop =
  match prop with
  | TL.Unsat f ->
    f ();
    env
  | TL.IsSubtype (ty1, ty2) -> sub_type_inner_helper env ~this_ty ty1 ty2
  | TL.IsEqual (ty1, ty2) ->
    (* These come only from invariant generics, with new_inference=false *)
    fst (Unify.unify env ty2 ty1)
  | TL.Conj props ->
    (* Evaluates list from left-to-right so preserves order of conjuncts *)
    List.fold_left
      ~init:env
      ~f:(process_simplify_subtype_result ~this_ty ~fail)
      props
  | TL.Disj [prop] ->
    process_simplify_subtype_result ~this_ty ~fail env prop

  | TL.Disj props ->
    let rec try_disj props =
      match props with
      | [] ->
        fail ();
        env
      | prop :: props ->
        Errors.try_ begin fun () ->
          process_simplify_subtype_result ~this_ty ~fail env prop end
          (fun _ -> try_disj props)
    in
      try_disj props

(* Attempt to "solve" a subtype assertion ty_sub <: ty_super.
 * Return a proposition that is equivalent, but simpler, than
 * the original assertion. Fail with Unsat error_function if
 * the assertion is unsatisfiable. Some examples:
 *   string <: arraykey  ==>  True    (represented as Conj [])
 * (For covariant C and a type variable v)
 *   C<string> <: C<v>   ==>  string <: v
 * (Assuming that C does *not* implement interface J)
 *   C <: J              ==>  Unsat _
 * (Assuming we have T <: D in tpenv, and class D implements I)
 *   vec<T> <: vec<I>    ==>  True
 * This last one would be left as T <: I if seen_generic_params is None
 *)
and simplify_subtype
  ~(seen_generic_params : SSet.t option)
  ?(deep : bool = true)
  ?(this_ty : locl ty option = None)
  (ty_sub : locl ty)
  (ty_super : locl ty)
  env : Env.env * TL.subtype_prop =
  log_subtype ~this_ty "simplify_subtype" env ty_sub ty_super;
  let env, ety_super = Env.expand_type env ty_super in
  let env, ety_sub = Env.expand_type env ty_sub in
  let again env acc ty_sub = (env, acc) &&&
      simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super in
  (* We *know* that the assertion is unsatisfiable *)
  let invalid_with f = env, TL.Unsat f in
  let invalid () =
    invalid_with (fun () ->
      TUtils.uerror (fst ety_super) (snd ety_super) (fst ety_sub) (snd ety_sub)) in
  (* We *know* that the assertion is valid *)
  let valid () = env, TL.valid in
  (* We don't know whether the assertion is valid or not *)
  let default () = env, TL.IsSubtype (ty_sub, ty_super) in
  let simplify_subtype_generic_sub name_sub opt_sub_cstr ty_super env =
  begin match seen_generic_params with
  | None -> default ()
  | Some seen ->
    (* If we've seen this type parameter before then we must have gone
     * round a cycle so we fail
     *)
    if SSet.mem name_sub seen
    then invalid ()
    else
    (* If the generic is actually an expression dependent type,
      we need to update this_ty
    *)
    let this_ty = if AbstractKind.is_generic_dep_ty name_sub &&
      Option.is_none this_ty then Some ety_sub else this_ty in
    let seen_generic_params = Some (SSet.add name_sub seen) in
    (* Otherwise, we collect all the upper bounds ("as" constraints) on
       the generic parameter, and check each of these in turn against
       ty_super until one of them succeeds
     *)
    let rec try_bounds tyl env =
      match tyl with
      | [] ->
        (* Try an implicit mixed = ?nonnull bound before giving up.
           This can be useful when checking T <: t, where type t is
           equivalent to but syntactically different from ?nonnull.
           E.g., if t is a generic type parameter T with nonnull as
           a lower bound.
         *)
        let r = Reason.Rimplicit_upper_bound (Reason.to_pos (fst ety_sub), "?nonnull") in
        let tmixed = (r, Toption (r, Tnonnull)) in
        env |>
        simplify_subtype ~seen_generic_params ~deep ~this_ty tmixed ty_super

      | [ty] ->
        env |>
        simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super

      | ty::tyl ->
        env |>
        try_bounds tyl |||
        simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super
    in
      env |>
      try_bounds (Option.to_list opt_sub_cstr
        @ Typing_set.elements (Env.get_upper_bounds env name_sub)) |>
      (* Turn error into a generic error about the type parameter *)
      if_unsat invalid
    end in

  let simplify_subtype_generic_super ty_sub name_super env =
    begin match seen_generic_params with
    | None -> default ()
    | Some seen ->

    (* If we've seen this type parameter before then we must have gone
     * round a cycle so we fail
     *)
    if SSet.mem name_super seen
    then invalid ()
    else
      let seen_generic_params = Some (SSet.add name_super seen) in
      (* Collect all the lower bounds ("super" constraints) on the
       * generic parameter, and check ty_sub against each of them in turn
       * until one of them succeeds *)
      let rec try_bounds tyl env =
        match tyl with
        | [] ->
          invalid ()

        | ty::tyl ->
          env |>
          simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty |||
          try_bounds tyl
      in
        (* Turn error into a generic error about the type parameter *)
        env |>
        try_bounds (Typing_set.elements (Env.get_lower_bounds env name_super)) |>
        if_unsat invalid
     end in

  match snd ety_sub, snd ety_super with
  | Tvar _, _ | _, Tvar _ ->
    if TypecheckerOptions.new_inference (Env.get_tcopt env)
    then default ()
    else assert false

  (* Internally, newtypes are always equipped with an upper bound.
   * In the case when no upper bound is specified in source code,
   * an implicit upper bound mixed = ?nonnull is added.
   *)
  | Tabstract (AKnewtype _, None), _
  | _, Tabstract (AKnewtype _, None) -> assert false

  | Terr, _ | _, Terr -> valid ()

  | (Tprim Nast.(Tint | Tbool | Tfloat | Tstring | Tresource | Tnum |
                 Tarraykey | Tnoreturn) |
     Tnonnull | Tfun _ | Ttuple _ | Tshape _ | Tabstract (AKenum _, _) |
     Tanon _ | Tobject | Tclass _ | Tarraykind _),
    Tnonnull ->
    valid ()
  | (Tdynamic | Toption _ | Tprim Nast.(Tnull | Tvoid) |
     Tabstract (AKdependent _, None)),
    Tnonnull ->
    invalid ()
  | Tabstract (AKnewtype _, Some ty), Tnonnull ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, Some _), Tnonnull ->
    default ()

  | Tdynamic, Tdynamic ->
    valid ()
  | (Tnonnull | Toption _ | Tprim _ | Tfun _ | Ttuple _ | Tshape _ |
     Tabstract (AKenum _, _) | Tanon _ | Tobject | Tclass _ | Tarraykind _),
    Tdynamic ->
    invalid ()
  | Tabstract (AKnewtype (_, _), Some ty), Tdynamic ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tdynamic ->
    default ()

  (* everything subtypes mixed *)
  | _, Toption (_, Tnonnull) -> valid ()
  (* null is the type of null and is a subtype of any option type. *)
  | Tprim Nast.Tnull, Toption _ -> valid ()
  (* void behaves with respect to subtyping as an abstract type with
   * an implicit upper bound ?nonnull
   *)
  | Tprim Nast.Tvoid, Toption ty_super' ->
    let r = Reason.Rimplicit_upper_bound (Reason.to_pos (fst ety_sub), "?nonnull") in
    let tmixed = (r, Toption (r, Tnonnull)) in
    env |>
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super' |||
    simplify_subtype ~seen_generic_params ~deep ~this_ty tmixed ty_super
  | Tdynamic, Toption ty_super ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super env
  (* ?ty_sub' <: ?ty_super' iff ty_sub' <: ?ty_super'. Reasoning:
   * If ?ty_sub' <: ?ty_super', then from ty_sub' <: ?ty_super' (widening) and transitivity
   * of <: it follows that ty_sub' <: ?ty_super'.  Conversely, if ty_sub' <: ?ty_super', then
   * by covariance and idempotence of ?, we have ?ty_sub' <: ??ty_sub' <: ?ty_super'.
   * Therefore, this step preserves the set of solutions.
   *)
  | Toption ty_sub', Toption _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub' ty_super env
  (* If ty_sub <: ?ty_super' and ty_sub does not contain null then we
   * must also have ty_sub <: ty_super'.  The converse follows by
   * widening and transitivity.  Therefore, this step preserves the set
   * of solutions.
   *)
  | (Tprim Nast.(Tint | Tbool | Tfloat | Tstring | Tresource | Tnum |
                 Tarraykey | Tnoreturn) |
     Tnonnull | Tfun _ | Ttuple _ | Tshape _ | Tanon _ | Tobject |
     Tclass _ | Tarraykind _ | Tabstract (AKenum _, _)),
    Toption ty_super' ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super' env
  | Tabstract (AKnewtype (name_sub, _), _),
    Toption (_, Tabstract (AKnewtype (name_super, _), _) as ty_super')
    when name_super = name_sub ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super' env
  (* If t1 <: ?t2 and t1 is an abstract type constrained as t1',
   * then t1 <: t2 or t1' <: ?t2.  The converse is obviously
   * true as well.  We can fold the case where t1 is unconstrained
   * into the case analysis below.
   *)
  | Tabstract ((AKnewtype _), Some ty), Toption arg_ty_super ->
    env |>
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub arg_ty_super |||
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super
  (* If t1 <: ?t2, where t1 is guaranteed not to contain null, then
   * t1 <: t2, and the converse is obviously true as well.
   *)
  | Tabstract (AKdependent _, None), Toption ty_super ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super env
  | Tabstract (AKgeneric name_sub, opt_sub_cstr), Toption arg_ty_super
    when Option.is_some seen_generic_params ->
    env |>
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub arg_ty_super |||
    (* Look up *)
    simplify_subtype_generic_sub name_sub opt_sub_cstr ty_super
  | Tabstract (AKdependent _, _), Toption _ -> default ()

  | Tprim (Nast.Tint | Nast.Tfloat), Tprim Nast.Tnum -> valid ()
  | Tprim (Nast.Tint | Nast.Tstring), Tprim Nast.Tarraykey -> valid ()
  | Tprim p1, Tprim p2 ->
    if p1 = p2 then valid () else invalid ()
  | Tabstract ((AKenum _), _), Tprim Nast.Tarraykey ->
    valid ()
  | (Tnonnull | Tdynamic | Tfun _ | Ttuple _ | Tshape _ | Tanon _ |
     Tabstract (AKenum _, None) | Tobject | Tclass _ | Tarraykind _),
    Tprim _ ->
    invalid ()
  | Toption _,
    Tprim Nast.(Tvoid | Tint | Tbool | Tfloat | Tstring | Tresource | Tnum |
                Tarraykey | Tnoreturn) ->
    invalid ()
  | Toption ty_sub', Tprim Nast.Tnull ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub' ty_super env
  | Tabstract ((AKnewtype _ | AKenum _), Some ty), Tprim _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tprim _ -> default ()

  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Ttuple _ | Tshape _ |
     Tabstract (AKenum _, _) | Tobject | Tclass _ | Tarraykind _), Tfun _ ->
    invalid ()
  | Tfun ft_sub, Tfun ft_super ->
    let r_sub, r_super = fst ety_sub, fst ety_super in
    simplify_subtype_funs ~seen_generic_params ~deep ~check_return:true
      r_sub ft_sub r_super ft_super env
  | Tanon (anon_arity, id), Tfun ft ->
    let r_sub, r_super = fst ety_sub, fst ety_super in
    begin match Env.get_anonymous env id with
      | None ->
        invalid_with (fun () -> Errors.anonymous_recursive_call (Reason.to_pos r_sub))
      | Some (reactivity, is_coroutine, ftys, _, anon) ->
        let p_super = Reason.to_pos r_super in
        let p_sub = Reason.to_pos r_sub in
        (* Add function type to set of types seen so far *)
        ftys := TUtils.add_function_type env ety_super !ftys;
        (env, TL.valid) |>
        check_with (subtype_reactivity env reactivity ft.ft_reactive
          || TypecheckerOptions.unsafe_rx (Env.get_tcopt env))
          (fun () -> Errors.fun_reactivity_mismatch
            p_super (TUtils.reactivity_to_string env reactivity)
            p_sub (TUtils.reactivity_to_string env ft.ft_reactive)) |>
        check_with (is_coroutine = ft.ft_is_coroutine) (fun () ->
          Errors.coroutinness_mismatch ft.ft_is_coroutine p_super p_sub) |>
        check_with (Unify.unify_arities
                  ~ellipsis_is_variadic:true anon_arity ft.ft_arity)
          (fun () -> Errors.fun_arity_mismatch p_super p_sub) |>
        (fun (env, prop) ->
          let env, _, ret = anon env ft.ft_params ft.ft_arity in
          (env, prop) &&&
          simplify_subtype ~seen_generic_params ~deep ~this_ty ret ft.ft_ret)
    end
  | Tabstract (AKnewtype _, Some ty), Tfun _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tfun _ -> default ()

  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Tfun _ | Tshape _ |
     Tabstract (AKenum _, _) | Tanon _ | Tobject | Tclass _ | Tarraykind _),
    Ttuple _ ->
    invalid ()
  (* (t1,...,tn) <: (u1,...,un) iff t1<:u1, ... , tn <: un *)
  | Ttuple tyl_sub, Ttuple tyl_super ->
    if List.length tyl_super = List.length tyl_sub
    then
      wfold_left2 (fun res ty_sub ty_super -> res
        &&& simplify_subtype ~seen_generic_params ~deep ty_sub ty_super)
        (env, TL.valid) tyl_sub tyl_super
    else invalid ()
  | Tabstract (AKnewtype _, Some ty), Ttuple _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Ttuple _ -> default ()

  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Tfun _ | Ttuple _ |
     Tabstract (AKenum _, _) | Tanon _ | Tobject | Tclass _ | Tarraykind _),
    Tshape _ ->
    invalid ()
  | Tshape (fields_known_sub, fdm_sub), Tshape (fields_known_super, fdm_super) ->
    let r_sub, r_super = fst ety_sub, fst ety_super in
      (**
       * shape_field_type A <: shape_field_type B iff:
       *   1. A is no more optional than B
       *   2. A's type <: B.type
       *)
      let on_common_field
          (env, acc) name
          { sft_optional = optional_super; sft_ty = ty_super }
          { sft_optional = optional_sub; sft_ty = ty_sub } =
        match optional_super, optional_sub with
          | true, _ | false, false ->
            (env, acc) &&& simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super
          | false, true ->
            (env, acc) |> with_error (fun () -> Errors.required_field_is_optional
              (Reason.to_pos r_sub)
              (Reason.to_pos r_super)
              (Env.get_shape_field_name name)) in
      let on_missing_omittable_optional_field res _ _ = res in
      let on_missing_non_omittable_optional_field
          res name { sft_ty = ty_super; _ } =
        let r = Reason.Rmissing_optional_field (
          Reason.to_pos r_sub,
          TUtils.get_printable_shape_field_name name
        ) in
        res &&& simplify_subtype ~seen_generic_params ~deep ~this_ty (r, TUtils.desugar_mixed r) ty_super in
      TUtils.apply_shape
        ~on_common_field
        ~on_missing_omittable_optional_field
        ~on_missing_non_omittable_optional_field
        ~on_error:(fun _ f -> invalid_with f)
        (env, TL.valid)
        (r_super, fields_known_super, fdm_super)
        (r_sub, fields_known_sub, fdm_sub)
  | Tabstract (AKnewtype _, Some ty), Tshape _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tshape _ -> default ()

  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Tfun _ | Ttuple _ | Tshape _ |
     Tabstract (AKenum _, None) | Tanon _ | Tobject | Tclass _ | Tarraykind _),
    Tabstract (AKnewtype _, _) ->
    invalid ()
  | Tabstract (AKnewtype (name_sub, tyl_sub), _),
    Tabstract (AKnewtype (name_super, tyl_super), _)
    when name_super = name_sub ->
      let td = Env.get_typedef env name_super in
      begin match td with
        | Some {td_tparams; _} ->
          let variancel = List.map td_tparams (fun (var,_,_,_) -> var) in
          simplify_subtype_variance ~seen_generic_params ~deep name_sub variancel tyl_sub tyl_super env
        | None ->
          default ()
      end
  | Tabstract ((AKnewtype _ | AKenum _), Some ty), Tabstract (AKnewtype _, _) ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tabstract (AKnewtype _, _) -> default ()

  | Tabstract (AKenum e_sub, _), Tabstract (AKenum e_super, _)
    when e_sub = e_super -> valid ()
  | Tclass ((_, class_name), _), Tabstract (AKenum enum_name, _)
    when enum_name = class_name -> valid ()
  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Tfun _ | Ttuple _ | Tshape _ |
     Tanon _ | Tobject | Tclass _ | Tarraykind _),
    Tabstract (AKenum _, _) ->
    invalid ()
  | Tabstract (AKenum _, None), Tabstract (AKenum _, _) -> invalid ()
  | Tabstract ((AKnewtype _ | AKenum _), Some ty), Tabstract (AKenum _, _) ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tabstract (AKenum _, _) -> default ()

  (* Primitives and other concrete types cannot be subtypes of dependent types *)
  | (Tnonnull | Tdynamic | Tprim _ | Tfun _ | Ttuple _ | Tshape _ |
     Tabstract (AKenum _, None) | Tanon _ | Tobject | Tarraykind _),
    Tabstract (AKdependent _, _) ->
    invalid ()
  | Tabstract (AKdependent d_sub, Some ty_sub),
    Tabstract (AKdependent d_super, Some ty_super)
    when d_sub = d_super ->
    (* Dependent types are identical but bound might be different *)
    let this_ty = Option.first_some this_ty (Some ety_sub) in
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super env
  | Tabstract ((AKnewtype _ | AKenum _), Some ty), Tabstract (AKdependent _, _) ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | (Toption _ | Tclass _ | Tabstract (AKdependent _, _)),
    Tabstract (AKdependent _, _) ->
    default ()

  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Ttuple _ | Tshape _ |
     Tabstract (AKenum _, _) | Tobject | Tclass _ | Tarraykind _),
    Tanon _ ->
    invalid ()
  | Tanon (_, id1), Tanon (_, id2) -> if id1 = id2 then valid () else invalid ()
  | Tabstract (AKnewtype _, Some ty), Tanon _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | (Tfun _ | Tabstract (AKdependent _, _)), Tanon _ -> default ()

  | Tobject, Tobject -> valid ()
  (* Any class type is a subtype of object *)
  | Tclass _, Tobject -> valid ()
  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Tfun _ | Ttuple _ | Tshape _ |
     Tabstract (AKenum _, _) | Tanon _ | Tarraykind _),
    Tobject ->
    invalid ()
  | Tabstract (AKnewtype _, Some ty), Tobject ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tabstract (AKdependent _, _), Tobject -> default ()

  | Tabstract (AKenum enum_name, None), Tclass ((_, class_name), _) ->
    if enum_name = class_name || class_name = SN.Classes.cXHPChild
    then valid ()
    else invalid ()
  | Tabstract (AKenum enum_name, Some ty), Tclass ((_, class_name), _) ->
    if enum_name = class_name
    then valid ()
    else simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tprim Nast.Tstring, Tclass ((_, class_name), _) ->
    if class_name = SN.Classes.cStringish || class_name = SN.Classes.cXHPChild
    then valid ()
    else invalid ()
  | Tprim Nast.(Tarraykey | Tint | Tfloat | Tnum), Tclass ((_, class_name), _) ->
    if class_name = SN.Classes.cXHPChild then valid () else invalid ()
  | (Tnonnull | Tdynamic | Tprim Nast.(Tnull | Tvoid | Tbool | Tresource | Tnoreturn) |
     Toption _ | Tfun _ | Ttuple _ | Tshape _ | Tanon _),
    Tclass _ ->
    invalid ()
  (* Match what's done in unify for non-strict code *)
  | Tobject, Tclass _ ->
    if Env.is_strict env then default () else valid ()
  | Tclass (x_sub, tyl_sub), Tclass (x_super, tyl_super) ->
    let p_sub, p_super = fst ety_sub, fst ety_super in
    let cid_super, cid_sub = (snd x_super), (snd x_sub) in
    (* This is side-effecting as it registers a dependency *)
    let class_def_sub = Env.get_class env cid_sub in
    if cid_super = cid_sub
    then
      (* We handle the case where a generic A<T> is used as A *)
      let tyl_super =
        if List.is_empty tyl_super && not (Env.is_strict env)
        then List.map tyl_sub (fun _ -> (p_super, Tany))
        else tyl_super in
      let tyl_sub =
        if List.is_empty tyl_sub && not (Env.is_strict env)
        then List.map tyl_super (fun _ -> (p_super, Tany))
        else tyl_sub in
      if List.length tyl_sub <> List.length tyl_super
      then begin
        let n_sub = String_utils.soi (List.length tyl_sub) in
        let n_super = String_utils.soi (List.length tyl_super) in
        invalid_with (fun () ->
          Errors.type_arity_mismatch (fst x_super) n_super (fst x_sub) n_sub)
      end
      else if List.is_empty tyl_sub && List.is_empty tyl_super
      then valid ()
      else
        let variancel =
          match class_def_sub with
          | None ->
            List.map tyl_sub (fun _ -> Ast.Invariant)
          | Some class_sub ->
            List.map class_sub.tc_tparams (fun (var, _, _, _) -> var) in

          (* C<t1, .., tn> <: C<u1, .., un> iff
           *   t1 <:v1> u1 /\ ... /\ tn <:vn> un
           * where vi is the variance of the i'th generic parameter of C,
           * and <:v denotes the appropriate direction of subtyping for variance v
           *)
          simplify_subtype_variance ~seen_generic_params ~deep cid_sub variancel tyl_sub tyl_super env
    else
      begin match class_def_sub with
      | None ->
        default ()

      | Some class_sub ->
        (* We handle the case where a generic A<T> is used as A *)
        let tyl_sub =
          if List.is_empty tyl_sub && not (Env.is_strict env)
          then List.map class_sub.tc_tparams (fun _ -> (p_sub, Tany))
          else tyl_sub in
        if List.length class_sub.tc_tparams <> List.length tyl_sub
        then
          invalid_with (fun () ->
          Errors.expected_tparam ~definition_pos:class_sub.tc_pos
            ~use_pos:(Reason.to_pos p_sub) (List.length class_sub.tc_tparams))
        else
          let ety_env =
          (* NOTE: We rely on the fact that we fold all ancestors of
           * ty_sub in its class_type so we will never hit this case
           * again. If this ever changes then we would need to store
           * ty_sub as the 'this_ty' in the uenv and be careful to
           * thread it through.
           *
           * This is covered by test/typecheck/this_tparam2.php
          *)
          {
            type_expansions = [];
            substs = Subst.make class_sub.tc_tparams tyl_sub;
            (* TODO: do we need this? *)
            this_ty = Option.value this_ty ~default:ty_sub;
            from_class = None;
            validate_dty = None;
          } in
            let up_obj = SMap.get cid_super class_sub.tc_ancestors in
            match up_obj with
              | Some up_obj ->
                let env, up_obj = Phase.localize ~ety_env env up_obj in
                simplify_subtype ~seen_generic_params ~deep ~this_ty up_obj ty_super env
              | None ->
                default ()
      end
  | Tabstract (AKnewtype _, Some ty), Tclass _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tarraykind _, Tclass ((_, class_name), _)
    when class_name = SN.Classes.cXHPChild -> valid ()
  | Tarraykind akind, Tclass ((_, coll), [tv_super])
    when (coll = SN.Collections.cTraversable ||
          coll = SN.Rx.cTraversable ||
          coll = SN.Collections.cContainer) ->
    let r = fst ety_sub in
      (match akind with
        (* array <: Traversable<t> and emptyarray <: Traversable<t> for any t *)
      | AKany -> valid ()
      | AKempty -> valid ()
      (* vec<tv> <: Traversable<tv_super>
       * iff tv <: tv_super
       * Likewise for vec<tv> <: Container<tv_super>
       *          and map<_,tv> <: Traversable<tv_super>
       *          and map<_,tv> <: Container<tv_super>
       *)
      | AKvarray tv
      | AKvec tv
      | AKdarray (_, tv)
      | AKvarray_or_darray tv
      | AKmap (_, tv) -> simplify_subtype ~seen_generic_params ~deep ~this_ty tv tv_super env
      | AKshape fdm ->
        Typing_arrays.fold_akshape_as_akmap_with_acc again env TL.valid r fdm
      | AKtuple fields ->
        Typing_arrays.fold_aktuple_as_akvec_with_acc again env TL.valid r fields
    )
  | Tarraykind akind, Tclass ((_, coll), [tk_super; tv_super])
    when (coll = SN.Collections.cKeyedTraversable
         || coll = SN.Rx.cKeyedTraversable
         || coll = SN.Collections.cKeyedContainer
         || coll = SN.Collections.cIndexish) ->
    let r = fst ety_sub in
      (match akind with
      | AKany -> valid ()
      | AKempty -> valid ()
      | AKvarray tv
      | AKvec tv ->
        env |>
        simplify_subtype ~seen_generic_params ~deep ~this_ty (r, Tprim Nast.Tint) tk_super &&&
        simplify_subtype ~seen_generic_params ~deep ~this_ty tv tv_super
      | AKvarray_or_darray tv ->
        let tk_sub =
          Reason.Rvarray_or_darray_key (Reason.to_pos r),
          Tprim Nast.Tarraykey in
        env |>
        simplify_subtype ~seen_generic_params ~deep ~this_ty tk_sub tk_super &&&
        simplify_subtype ~seen_generic_params ~deep ~this_ty tv tv_super
      | AKdarray (tk, tv)
      | AKmap (tk, tv) ->
        env |>
        simplify_subtype ~seen_generic_params ~deep ~this_ty tk tk_super &&&
        simplify_subtype ~seen_generic_params ~deep ~this_ty tv tv_super
      | AKshape fdm ->
        Typing_arrays.fold_akshape_as_akmap_with_acc again env TL.valid r fdm
      | AKtuple fields ->
        Typing_arrays.fold_aktuple_as_akvec_with_acc again env TL.valid r fields
      )
  | Tarraykind _, Tclass _ -> invalid ()
  | Tabstract (AKdependent _, _), Tclass _ -> default ()

  (* Arrays *)
  | Ttuple _, Tarraykind AKany ->
    if TypecheckerOptions.disallow_array_as_tuple (Env.get_tcopt env)
    then invalid ()
    else valid ()
  | (Tnonnull | Tdynamic | Toption _ | Tprim _ | Tfun _ | Ttuple _ |
     Tshape _ | Tabstract (AKenum _, _) | Tanon _ | Tobject | Tclass _),
    Tarraykind _ ->
    invalid ()
  | Tabstract (AKnewtype _, Some ty), Tarraykind _ ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super env
  | Tarraykind ak_sub, Tarraykind ak_super ->
    let r = fst ety_sub in
    begin match ak_sub, ak_super with
    (* An array of any kind is a subtype of an array of AKany *)
    | _, AKany ->
      valid ()

    (* An empty array is a subtype of any array type *)
    | AKempty, _ ->
      valid ()

    (* array is a subtype of varray_or_darray<_> *)
    | AKany, AKvarray_or_darray (_, Tany) ->
      valid ()

    | AKany, _ ->
      let safe_array = TypecheckerOptions.safe_array (Env.get_tcopt env) in
      if safe_array then invalid () else valid ()

    | AKshape fdm, (
        AKvarray _
      | AKvec _
      | AKdarray _
      | AKvarray_or_darray _
      | AKmap _) ->
      Typing_arrays.fold_akshape_as_akmap_with_acc
       begin fun env acc ty ->
         (env, acc) &&& simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super
       end env TL.valid r fdm

    | AKtuple fields,
      (
        AKvarray _
      | AKvec _
      | AKdarray _
      | AKvarray_or_darray _
      | AKmap _
      ) ->
      Typing_arrays.fold_aktuple_as_akvec_with_acc
       begin fun env acc ty ->
         (env, acc) &&& simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super
       end env TL.valid r fields

    (* varray_or_darray<ty1> <: varray_or_darray<ty2> iff t1 <: ty2
       But, varray_or_darray<ty1> is never a subtype of a vect-like array *)
    | AKvarray_or_darray ty_sub, AKvarray_or_darray ty_super ->
      simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super env

    | (AKvarray ty_sub | AKvec ty_sub),
      (AKvarray ty_super | AKvec ty_super | AKvarray_or_darray ty_super) ->
      simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super env

    | (AKdarray (tk_sub, tv_sub) | AKmap (tk_sub, tv_sub)),
      (AKdarray (tk_super, tv_super) | AKmap (tk_super, tv_super)) ->
      env |>
      simplify_subtype ~seen_generic_params ~deep ~this_ty tk_sub tk_super &&&
      simplify_subtype ~seen_generic_params ~deep ~this_ty tv_sub tv_super

    | (AKdarray (tk_sub, tv_sub) | AKmap (tk_sub, tv_sub)),
      (AKvarray_or_darray tv_super) ->
      let tk_super =
        Reason.Rvarray_or_darray_key (Reason.to_pos (fst ety_super)),
        Tprim Nast.Tarraykey in
      env |>
      simplify_subtype ~seen_generic_params ~deep ~this_ty tk_sub tk_super &&&
      simplify_subtype ~seen_generic_params ~deep ~this_ty tv_sub tv_super

    | (AKvarray elt_ty | AKvec elt_ty), (AKdarray _ | AKmap _)
        when not (TypecheckerOptions.safe_vector_array (Env.get_tcopt env)) ->
          let int_reason = Reason.Ridx (Reason.to_pos r, Reason.Rnone) in
          let int_type = int_reason, Tprim Nast.Tint in
          simplify_subtype ~seen_generic_params ~deep ~this_ty
            (r, Tarraykind (AKmap (int_type, elt_ty))) ty_super env

    (* any other array subtyping is unsatisfiable *)
    | _ ->
      invalid ()
    end
  | Tabstract (AKdependent _, _), Tarraykind _ -> default ()

  (* ty_sub <: union{ty_super'} iff ty_sub <: ty_super' *)
  | _, Tunresolved [ty_super'] when deep ->
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super' env

  (* Don't yet attempt to deal with unresolved types *)
  | Tunresolved _, _
  | _, Tunresolved _ ->
    default ()

  (* This is sort of a hack because our handling of Toption is highly
   * dependent on how the type is structured. When we see a bare
   * dependent type we strip it off at this point since it shouldn't be
   * relevant to subtyping any more.
   *)

  | Tabstract (AKdependent (`expr _, []), Some ty_sub), _ ->
    let this_ty = Option.first_some this_ty (Some ety_sub) in
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty_sub ty_super env

  | Tany, _ -> default ()
  | _, Tany -> default ()

  (* If subtype and supertype are the same generic parameter, we're done *)
  | Tabstract (AKgeneric name_sub, _), Tabstract (AKgeneric name_super, _)
       when name_sub = name_super
    -> valid ()

  (* Supertype is generic parameter *and* subtype is a newtype with bound.
   * We need to make this a special case because there is a *choice*
   * of subtyping rule to apply. See details in the case of dependent type
   * against generic parameter which is similar
   *)
  | Tabstract (AKnewtype (_, _), Some ty), Tabstract (AKgeneric name_super, _)
    when Option.is_some seen_generic_params ->
    env |>
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super |||
    simplify_subtype_generic_super ty_sub name_super

  (* void behaves with respect to subtyping as an abstract type with
   * an implicit upper bound ?nonnull
   *)
  | Tprim Nast.Tvoid, Tabstract (AKgeneric name_super, _)
    when Option.is_some seen_generic_params ->
    let r = Reason.Rimplicit_upper_bound (Reason.to_pos (fst ety_sub), "?nonnull") in
    let tmixed = (r, Toption (r, Tnonnull)) in
    env |>
    simplify_subtype ~seen_generic_params ~deep ~this_ty tmixed ty_super |||
    simplify_subtype_generic_super ty_sub name_super

  (* Supertype is generic parameter *and* subtype is dependent.
   * We need to make this a special case because there is a *choice*
   * of subtyping rule to apply.
   *
   * Example. First suppose that we have the definition
   *
   *     abstract const type TC as C
   *
   * (1) Now suppose we have to check
   *       this::TC <: Tu
   *     where we have the constraint
   *       <Tu super C>.
   *     Then it's necessary to apply the rule for AKdependent first, so we
   *     reduce this problem to
   *       C <: Tu
   *     and then call sub_generic_params to deal with the type parameter.
   * (2) Alternatively, suppose we again have to check
   *       this::TC <: Tu
   *     but this time we have the constraint
   *       <Tu super this::TC>.
   *    Then if we first reduce the problem to C <: Tu we fail;
   *    but if we also try sub_generic_params then we succeed, because
   *    we end up checking this::TC <: this::TC.
  *)
  | Tabstract (AKdependent _, Some ty), Tabstract (AKgeneric name_super, _)
    when Option.is_some seen_generic_params ->
    env |>
    simplify_subtype_generic_super ty_sub name_super |||
    (let this_ty = Option.first_some this_ty (Some ety_sub) in
    simplify_subtype ~seen_generic_params ~deep ~this_ty ty ty_super)

  (* Subtype or supertype is generic parameter
   * We delegate these cases to a separate function in order to catch cycles
   * in constraints e.g. <T1 as T2, T2 as T3, T3 as T1>
   *)
  | Tabstract (AKgeneric name_sub, opt_sub_cstr), _ ->
    simplify_subtype_generic_sub name_sub opt_sub_cstr ty_super env

  | _, Tabstract (AKgeneric name_super, _) ->
    simplify_subtype_generic_super ty_sub name_super env

and simplify_subtype_variance
  ~(seen_generic_params : SSet.t option)
  ~(deep : bool)
  (cid : string)
  (variancel : Ast.variance list)
  (children_tyl : locl ty list)
  (super_tyl : locl ty list)
  : Env.env -> Env.env * TL.subtype_prop
  = fun env ->
  match variancel, children_tyl, super_tyl with
  | [], _, _
  | _, [], _
  | _, _, [] -> valid env
  | variance :: variancel, child :: childrenl, super :: superl ->
      begin match variance with
      | Ast.Covariant ->
        simplify_subtype ~seen_generic_params ~deep ~this_ty:None child super env
      | Ast.Contravariant ->
        let super = (Reason.Rcontravariant_generic (fst super,
          Utils.strip_ns cid), snd super) in
        simplify_subtype ~seen_generic_params ~deep ~this_ty:None super child env
      | Ast.Invariant ->
        let super' = (Reason.Rinvariant_generic (fst super,
          Utils.strip_ns cid), snd super) in
        if deep
        then
          env |>
          simplify_subtype ~seen_generic_params ~deep ~this_ty:None child super' &&&
          simplify_subtype ~seen_generic_params ~deep ~this_ty:None super child
        else
          env, TL.IsEqual (child, super')
      end &&&
      simplify_subtype_variance ~seen_generic_params ~deep cid variancel childrenl superl

and simplify_subtype_params
  ~seen_generic_params
  ~deep
  ?(is_method : bool = false)
  ?(check_params_reactivity = false)
  ?(check_params_mutability = false)
  (subl : locl fun_param list)
  (superl : locl fun_param list)
  (variadic_sub_ty : locl ty option)
  (variadic_super_ty : locl ty option)
  env =

  match subl, superl with
  (* When either list runs out, we still have to typecheck that
  the remaining portion sub/super types with the other's variadic.
  For example, if
  ChildClass {
    public function a(int $x = 0, string ... $args) // superl = [int], super_var = string
  }
  overrides
  ParentClass {
    public function a(string ... $args) // subl = [], sub_var = string
  }
  , there should be an error because the first argument will be checked against
  int, not string that is, ChildClass::a("hello") would crash,
  but ParentClass::a("hello") wouldn't.

  Similarly, if the other list is longer, aka
  ChildClass  extends ParentClass {
    public function a(mixed ... $args) // superl = [], super_var = mixed
  }
  overrides
  ParentClass {
    //subl = [string], sub_var = string
    public function a(string $x = 0, string ... $args)
  }
  It should also check that string is a subtype of mixed.
  *)
  | [], _ -> (match variadic_super_ty with
    | None -> valid env
    | Some ty -> simplify_supertype_params_with_variadic ~seen_generic_params ~deep superl ty env)
  | _, [] -> (match variadic_sub_ty with
    | None -> valid env
    | Some ty -> simplify_subtype_params_with_variadic ~seen_generic_params ~deep subl ty env)
  | sub :: subl, super :: superl ->
    env |>
    begin
      if check_params_reactivity
      then subtype_fun_params_reactivity sub super
      else valid
    end &&&
    begin
      if check_params_mutability
      then check_mutability
        ~is_receiver:false
        sub.fp_pos sub.fp_mutability super.fp_pos super.fp_mutability
      else valid
    end &&&

    fun env ->
    begin
      let { fp_type = ty_sub; _ } = sub in
      let { fp_type = ty_super; _ } = super in
      (* Check that the calling conventions of the params are compatible.
       * We don't currently raise an error for reffiness because function
       * hints don't support '&' annotations (enforce_ctpbr = false). *)
      Unify.unify_param_modes ~enforce_ctpbr:is_method sub super;
      Unify.unify_accept_disposable sub super;
      let env = { env with Env.pos = Reason.to_pos (fst ty_sub) } in
      match sub.fp_kind, super.fp_kind with
      | FPinout, FPinout ->
        (* Inout parameters are invariant wrt subtyping for function types. *)
        env |>
        simplify_subtype ~seen_generic_params ~deep ty_super ty_sub &&&
        simplify_subtype ~seen_generic_params ~deep ty_sub ty_super
      | _ ->
        env |>
        simplify_subtype ~seen_generic_params ~deep ty_sub ty_super
      end &&&
    simplify_subtype_params ~seen_generic_params ~deep ~is_method subl superl
      variadic_sub_ty variadic_super_ty

and simplify_subtype_params_with_variadic
  ~seen_generic_params
  ~deep
  (subl : locl fun_param list)
  (variadic_ty : locl ty)
  env =
  match subl with
  | [] -> valid env
  | { fp_type = sub; _ } :: subl ->
    let env = { env with Env.pos = Reason.to_pos (fst sub) } in
    env |>
    simplify_subtype ~seen_generic_params ~deep sub variadic_ty &&&
    simplify_subtype_params_with_variadic ~seen_generic_params ~deep subl variadic_ty

and simplify_supertype_params_with_variadic
  ~seen_generic_params
  ~deep
  (superl : locl fun_param list)
  (variadic_ty : locl ty)
  env =
  match superl with
  | [] -> valid env
  | { fp_type = super; _ } :: superl ->
    let env = { env with Env.pos = Reason.to_pos (fst super) } in
    env |>
    simplify_subtype ~seen_generic_params ~deep variadic_ty super &&&
    simplify_supertype_params_with_variadic ~seen_generic_params ~deep superl variadic_ty

and subtype_reactivity
  ?(extra_info: reactivity_extra_info option)
  ?(is_call_site = false)
  (env : Env.env)
  (r_sub : reactivity)
  (r_super : reactivity) : bool =

  let maybe_localize t =
    match t with
    | DeclTy t ->
      let ety_env = Phase.env_with_self env in
      let _, t = Phase.localize ~ety_env env t in
      t
    | LoclTy t -> t in

  let parent_class_ty =
    Option.bind extra_info (fun { parent_class_ty = parent_cls; _ } ->
      Option.map parent_cls ~f:maybe_localize) in

  (* for method declarations check if condition type for r_super includes
     reactive method with a matching name. If yes - then it will act as a guarantee
     that derived class will have to redefine the method with a shape required
     by condition type (reactivity of redefined method must be subtype of reactivity
     of method in interface) *)
  let condition_type_has_matching_reactive_method condition_type_super (method_name, is_static) =
    let m =
      ConditionTypes.try_get_method_from_condition_type
        env condition_type_super is_static method_name in
    match m with
    | Some { ce_type = lazy (_, Tfun f); _  } ->
      (* check that reactivity of interface method (effectively a promised
         reactivity of a method in derived class) is a subtype of r_super.
         NOTE: we check only for unconditional reactivity since conditional
         version does not seems to yield a lot and will requre implementing
         cycle detection for condition types *)
      begin match f.ft_reactive with
      | Reactive None | Shallow None | Local None ->
        let extra_info = {
          empty_extra_info with parent_class_ty = Some (DeclTy condition_type_super)
        } in
        subtype_reactivity ~extra_info env f.ft_reactive r_super
      | _ -> false
      end
    | _ -> false in
  match r_sub, r_super, extra_info with
  (* anything is a subtype of nonreactive functions *)
  | _, Nonreactive, _ -> true
  (* to compare two maybe reactive values we need to unwrap them *)
  | MaybeReactive sub, MaybeReactive super, _ ->
    subtype_reactivity ?extra_info ~is_call_site env sub super
  (* for explicit checks at callsites implicitly unwrap maybereactive value:
     function f(<<__OnlyRxIfRxFunc>> F $f)
     f(<<__RxLocal>> () ==> {... })
     here parameter will be maybereactive and argument - rxlocal
     *)
  | sub, MaybeReactive super, _ when is_call_site ->
    subtype_reactivity ?extra_info ~is_call_site env sub super
  (* if is_call_site is falst ignore maybereactive flavors.
     This usually happens during subtype checks for arguments and when target
     function is conditionally reactive we'll do the proper check
     in typing_reactivity.check_call. *)
  | _, MaybeReactive _, _ when not is_call_site -> true
  (* ok:
    class A { function f((function(): int) $f) {} }
    class B extends A {
      <<__Rx>>
      function f(<<__AtMostRxAsFunc>> (function(): int) $f);
    }
    reactivity for arguments is checked contravariantly *)
  | _, RxVar None, _
  (* ok:
     <<__Rx>>
     function f(<<__AtMostRxAsFunc>> (function(): int) $f) { return $f() }  *)
  | RxVar None, RxVar _, _ -> true
  | RxVar (Some sub), RxVar (Some super), _
  | sub, RxVar (Some super), _ ->
    subtype_reactivity ?extra_info ~is_call_site env sub super
  | RxVar _, _, _ -> false
  | (Local cond_sub | Shallow cond_sub | Reactive cond_sub), Local cond_super, _
  | (Shallow cond_sub | Reactive cond_sub), Shallow cond_super, _
  | Reactive cond_sub, Reactive cond_super, _
    when subtype_param_rx_if_impl env cond_super parent_class_ty cond_sub ->
    true
  (* function type TSub of method M with arbitrary reactivity in derive class
     can be subtype of conditionally reactive function type TSuper of method M
     defined in base class when condition type has reactive method M.
     interface Rx {
       <<__Rx>>
       public function f(): int;
     }
     class A {
       <<__RxIfImplements(Rx::class)>>
       public function f(): int { ... }
     }
     class B extends A {
       public function f(): int { ... }
     }
     This should be OK because:
     - B does not implement Rx (B::f is not compatible with Rx::f) which means
     that calling ($b : B)->f() will not be treated as reactive
     - if one of subclasses of B will decide to implement B - they will be forced
     to redeclare f which now will shadow B::f. Note that B::f will still be
     accessible as parent::f() but will be treated as non-reactive call.
     *)
  | _, (Reactive (Some t) | Shallow (Some t) | Local (Some t)), Some { method_info = Some mi; _ }
    when condition_type_has_matching_reactive_method t mi ->
    true
  (* call_site specific cases *)
  (* shallow can call into local *)
  | Local cond_sub, Shallow cond_super, _ when
    is_call_site &&
    subtype_param_rx_if_impl env cond_super parent_class_ty cond_sub ->
    true
  (* local can call into non-reactive *)
  | Nonreactive, Local _, _ when is_call_site -> true
  | _ -> false

and should_check_fun_params_reactivity
  (ft_super: locl fun_type) = ft_super.ft_reactive <> Nonreactive

(* checks condition described by OnlyRxIfImpl condition on parameter is met  *)
and subtype_param_rx_if_impl
  (env: Env.env)
  (cond_type_sub: decl ty option)
  (declared_type_sub: locl ty option)
  (cond_type_super: decl ty option) =
  let cond_type_sub =
    Option.map cond_type_sub ~f:(ConditionTypes.localize_condition_type env) in
  let cond_type_super =
    Option.map cond_type_super ~f:(ConditionTypes.localize_condition_type env) in
  match cond_type_sub, cond_type_super with
  (* no condition types - do nothing *)
  | None, None -> true
  (* condition type is specified only for subtype - ok
    Consider this case:
    Receivers here are treated as parameters and checked contravariantly
    abstract class A {
      <<__RxLocal, __OnlyRxIfImpl(Rx1::class)>>
      public abstract function condlocalrx(): int;
    }
    abstract class B extends A {
      // ok to override cond local with local (if condition is not met - method
      // in base class is non-reactive )
      <<__Override, __RxLocal>>
      public function condlocalrx(): int {
        return 1;
      }
    }
  *)
  | Some _, None -> true
  (* condition types are set for both sub and super types *)
  | Some cond_type_sub, Some cond_type_super ->
    is_sub_type env cond_type_sub cond_type_super
  (* condition type is set for super type, check if declared type of
     subtype is a subtype of condition type
     Receivers here are treated as parameters and checked contravariantly
     interface Rx {
       <<__Rx>>
       public function f(int $a): void;
     }
     class A<T> {
       <<__Rx, __OnlyRxIfImpl(Rx::class)>>
       public function f(T $a): void {
       }
     }
     // B <: Rx so B::f is completely reactive
     class B extends A<int> implements Rx {
     } *)
  | None, Some cond_type_super ->
    Option.value_map declared_type_sub
      ~default:false
      ~f:(fun declared_type_sub -> is_sub_type env declared_type_sub cond_type_super)

(* checks reactivity conditions for function parameters *)
and subtype_fun_params_reactivity
  (p_sub: locl fun_param)
  (p_super: locl fun_param)
  env =
  match p_sub.fp_rx_annotation, p_super.fp_rx_annotation with
  (* no conditions on parameters - do nothing *)
  | None, None -> valid env
  (* both parameters are conditioned to be rx function - no need to check anything *)
  | Some Param_rx_var, Some Param_rx_var -> valid env
  | None, Some Param_rx_var ->
    (* parameter is conditionally reactive in supertype and missing condition
      in subtype - this is ok only if parameter in subtype is reactive
      <<__Rx>>
      function super((function(): int) $f);
      <<__Rx>>
      function sub(<<__OnlyRxIfRxFunc>> (function(): int) $f);
      We check if sub <: super. parameters are checked contravariantly
      so we need to verify that
      (function(): int) $f <: <<__OnlyRxIfRxFunc>> (function(): int) $f

      Suppose this is legal, then this will be allowed (in pseudo-code)

      function sub(<<__OnlyRxIfRxFunc>> (function(): int) $f) {
        $f(); // can call function here since it is conditionally reactive
      }
      <<__Rx>>
      function g() {
        $f: super = sub;
        // invoke non-reactive code in reactive context which is bad
        $f(() ==> { echo 1; });
      }
    }
    It will be safe if parameter in super will be completely reactive,
    hence check below *)
    let _, p_sub_type = Env.expand_type env p_sub.fp_type in
    begin match p_sub_type with
    | _, Tfun tfun when tfun.ft_reactive <> Nonreactive -> valid env
    | _, Tfun _ ->
      env, TL.Unsat (fun () -> Errors.rx_parameter_condition_mismatch
        SN.UserAttributes.uaAtMostRxAsFunc p_sub.fp_pos p_super.fp_pos)
    (* parameter type is not function - error will be reported in different place *)
    | _ -> valid env
    end
  | cond_sub, cond_super ->
    let cond_type_sub =
      match cond_sub with
      | Some (Param_rx_if_impl t) -> Some t
      | _ -> None in
    let cond_type_super =
      match cond_super with
      | Some (Param_rx_if_impl t) -> Some t
      | _ -> None in
    let ok =
      subtype_param_rx_if_impl env cond_type_sub (Some p_sub.fp_type) cond_type_super in
      check_with ok (fun () ->
      Errors.rx_parameter_condition_mismatch
        SN.UserAttributes.uaOnlyRxIfImpl p_sub.fp_pos p_super.fp_pos) (env, TL.valid)

(* Helper function for subtyping on function types: performs all checks that
 * don't involve actual types:
 *   <<__ReturnDisposable>> attribute
 *   <<__MutableReturn>> attribute
 *   <<__Rx>> attribute
 *   <<__Mutable>> attribute
 *   whether or not the function is a coroutine
 *   variadic arity
 *)
and check_subtype_funs_attributes
  ?(extra_info: reactivity_extra_info option)
  (r_sub : Reason.t)
  (ft_sub : locl fun_type)
  (r_super : Reason.t)
  (ft_super : locl fun_type) env =
  let p_sub = Reason.to_pos r_sub in
  let p_super = Reason.to_pos r_super in
  (env, TL.valid) |>
  check_with
    (subtype_reactivity ?extra_info env ft_sub.ft_reactive ft_super.ft_reactive)
    (fun () -> Errors.fun_reactivity_mismatch
      p_super (TUtils.reactivity_to_string env ft_super.ft_reactive)
      p_sub (TUtils.reactivity_to_string env ft_sub.ft_reactive))
  |>
  check_with
    (ft_sub.ft_is_coroutine = ft_super.ft_is_coroutine)
    (fun () -> Errors.coroutinness_mismatch ft_super.ft_is_coroutine p_super p_sub) |>
  check_with
    (ft_sub.ft_return_disposable = ft_super.ft_return_disposable)
    (fun () -> Errors.return_disposable_mismatch ft_super.ft_return_disposable p_super p_sub) |>
  (* it is ok for subclass to return mutably owned value and treat it as immutable -
  the fact that value is mutably owned guarantees it has only single reference so
  as a result this single reference will be immutable. However if super type
  returns mutable value and subtype yields immutable value - this is not safe.
  NOTE: error is not reported if child is non-reactive since it does not have
  immutability-by-default behavior *)
  check_with
    (ft_sub.ft_returns_mutable = ft_super.ft_returns_mutable
    || not ft_super.ft_returns_mutable
    || ft_sub.ft_reactive = Nonreactive)
    (fun () -> Errors.mutable_return_result_mismatch ft_super.ft_returns_mutable p_super p_sub) |>
  check_with
    (ft_super.ft_reactive = Nonreactive
    || ft_super.ft_returns_void_to_rx
    || not ft_sub.ft_returns_void_to_rx)
    (fun () ->
    (*  __ReturnsVoidToRx can be omitted on subtype, in this case using subtype
       via reference to supertype in rx context will be ok since result will be
       discarded. The opposite is not true:
       class A {
         <<__Rx, __Mutable>>
         public function f(): A { return new A(); }
       }
       class B extends A {
         <<__Rx, __Mutable, __ReturnsVoidToRx>>
         public function f(): A { return $this; }
       }

       <<__Rx, __MutableReturn>>
       function f(): A { return new B(); }
       $a = HH\Rx\mutable(f());
       $a1 = $a->f(); // immutable alias to mutable reference *)
      Errors.return_void_to_rx_mismatch ~pos1_has_attribute:true p_sub p_super) |>
  (* check mutability only for reactive functions *)
  let check_params_mutability =
    ft_super.ft_reactive <> Nonreactive &&
    ft_sub.ft_reactive <> Nonreactive in
  fun (env, prop) -> (if check_params_mutability
  (* check mutability of receivers *)
  then (env, prop) &&&
    check_mutability ~is_receiver:true
      p_super ft_super.ft_mutability p_sub ft_sub.ft_mutability else env, prop) |>
  check_with
    ((arity_min ft_sub.ft_arity) <= (arity_min ft_super.ft_arity))
    (fun () -> Errors.fun_too_many_args p_sub p_super)
    |>
  fun res -> (match ft_sub.ft_arity, ft_super.ft_arity with
    | Fellipsis _, Fvariadic _ ->
      (* The HHVM runtime ignores "..." entirely, but knows about
       * "...$args"; for contexts for which the runtime enforces method
       * compatibility (currently, inheritance from abstract/interface
       * methods), letting "..." override "...$args" would result in method
       * compatibility errors at runtime. *)
      with_error (fun () -> Errors.fun_variadicity_hh_vs_php56 p_sub p_super) res
    | Fstandard (_, sub_max), Fstandard (_, super_max) ->
      if sub_max < super_max
      then with_error (fun () -> Errors.fun_too_few_args p_sub p_super) res else res
    | Fstandard _, _ -> with_error (fun () -> Errors.fun_unexpected_nonvariadic p_sub p_super) res
    | _, _ -> res
  )

(* This implements basic subtyping on non-generic function types:
 *   (1) return type behaves covariantly
 *   (2) parameter types behave contravariantly
 *   (3) special casing for variadics, and various reactivity and mutability attributes
 *)
 and simplify_subtype_funs
  ~seen_generic_params
  ~deep
  ~(check_return : bool)
  ?(extra_info: reactivity_extra_info option)
  (r_sub : Reason.t)
  (ft_sub : locl fun_type)
  (r_super : Reason.t)
  (ft_super : locl fun_type)
  env
   : Env.env * TL.subtype_prop =
  let variadic_subtype = match ft_sub.ft_arity with
    | Fvariadic (_, {fp_type = var_sub; _ }) -> Some var_sub
    | _ -> None in
  let variadic_supertype =  match ft_super.ft_arity with
    | Fvariadic (_, {fp_type = var_super; _ }) -> Some var_super
    | _ -> None in

  (* First apply checks on attributes, coroutine-ness and variadic arity *)
  env |>
  check_subtype_funs_attributes ?extra_info r_sub ft_sub r_super ft_super &&&

  (* Now do contravariant subtyping on parameters *)
  begin
    match variadic_subtype, variadic_supertype with
    | Some var_sub, Some var_super -> simplify_subtype ~seen_generic_params ~deep var_super var_sub
    | _ -> valid
  end &&&

  begin
    let check_params_mutability =
      ft_super.ft_reactive <> Nonreactive &&
      ft_sub.ft_reactive <> Nonreactive in
    let is_method =
      (Option.map extra_info (fun i -> Option.is_some i.method_info)) = Some true in
    simplify_subtype_params
      ~seen_generic_params
      ~deep
      ~is_method
      ~check_params_reactivity:(should_check_fun_params_reactivity ft_super)
      ~check_params_mutability
      ft_super.ft_params ft_sub.ft_params variadic_subtype variadic_supertype
  end &&&

  (* Finally do covariant subtryping on return type *)
  if check_return
  then simplify_subtype ~seen_generic_params ~deep ft_sub.ft_ret ft_super.ft_ret
  else valid

(* One of the main entry points to this module *)
and sub_type
  (env : Env.env)
  (ty_sub : locl ty)
  (ty_super : locl ty) : Env.env =
    sub_type_inner env ~this_ty:None ty_sub ty_super

and props_to_env env remain props =
  match props with
  | [] ->
    env, List.rev remain
  | TL.IsSubtype (((r, Tvar var) as ty'), ty) :: props ->
    Typing_log.(log_with_level env "prop" 2 (fun () ->
    log_types (Reason.to_pos r) env
      [Log_head ("Typing_subtype.props_to_env/Env.add_tyvar_upper_bound",
      [Log_type ("var", ty'); Log_type ("ty", ty)])]));
    props_to_env (Env.add_tyvar_upper_bound env var ty) remain props
  | TL.IsSubtype (ty, ((r, Tvar var) as ty')) :: props ->
    Typing_log.(log_with_level env "prop" 2 (fun () ->
    log_types (Reason.to_pos r) env
      [Log_head ("Typing_subtype.props_to_env/Env.add_tyvar_lower_bound",
      [Log_type ("var", ty'); Log_type ("ty", ty)])]));
    props_to_env (Env.add_tyvar_lower_bound env var ty) remain props
  | TL.Conj props' :: props ->
    props_to_env env remain (props' @ props)
  | prop :: props ->
    props_to_env env (prop::remain) props

(* Move any top-level conjuncts of the form Tvar v <: t or t <: Tvar v to
 * the type variable environment. To do: use intersection and union to
 * simplify bounds.
 *)
and prop_to_env env prop =
  match prop with
  | TL.Conj props ->
    let env, props' = props_to_env env [] props in
    env, TL.conj_list props'
  | TL.IsSubtype _ ->
    let env, props' = props_to_env env [] [prop] in
    env, TL.conj_list props'
  | _ ->
    env, prop

and sub_type_inner
  (env : Env.env)
  ~(this_ty : locl ty option)
  (ty_sub: locl ty)
  (ty_super: locl ty) : Env.env =
  log_subtype ~this_ty "sub_type_inner" env ty_sub ty_super;
  let new_inference = TypecheckerOptions.new_inference (Env.get_tcopt env) in
  let env, prop = simplify_subtype
    ~seen_generic_params:empty_seen ~deep:new_inference ~this_ty ty_sub ty_super env in
  let env, prop = if new_inference then prop_to_env env prop else env, prop in
  let env =
    if new_inference || TypecheckerOptions.experimental_feature_enabled
         (Env.get_tcopt env)
         TypecheckerOptions.experimental_track_subtype_prop
    then begin
      log_prop env prop;
      Typing_log.log_prop 2 env.Env.pos "sub_type_inner: add_subtype_prop" env prop;
      Env.add_subtype_prop env prop
    end
    else env in
  process_simplify_subtype_result ~this_ty
    ~fail:(fun () -> TUtils.uerror (fst ty_super) (snd ty_super) (fst ty_sub) (snd ty_sub))
    env
    prop

(* Deal with the cases not dealt with by simplify_subtype *)
and sub_type_inner_helper env ~this_ty
  ty_sub ty_super =
  let env, ety_super =
    Env.expand_type env ty_super in
  let env, ety_sub =
    Env.expand_type env ty_sub in
  (* Default error *)

  let fail () =
    TUtils.uerror (fst ety_super) (snd ety_super) (fst ety_sub) (snd ety_sub);
    env in

  log_subtype ~this_ty "sub_type_inner_helper" env ty_sub ty_super;

  match ety_sub, ety_super with
  | (_, Terr), _
  | _, (_, Terr) -> env

  | (_, Tunresolved _), (_, Tunresolved _) ->
    fst (Unify.unify env ty_super ty_sub)

(****************************************************************************)
(* ### Begin Tunresolved madness ###
 * If the supertype is a
 * Tunresolved, we allow it to keep growing, which is the desired behavior for
 * e.g. figuring out the type of a generic, but if the subtype is a
 * Tunresolved, then we check that all the members are indeed subtypes of the
 * given supertype, which is the desired behavior for e.g. checking function
 * return values. In general, if a supertype is Tunresolved, then we
 * consider it to be "not yet finalized", but if a subtype is Tunresolved
 * and the supertype isn't, we've probably hit a type annotation
 * and should consider the supertype to be definitive.
 *
 * However, sometimes we want this behavior reversed, e.g. when the type
 * annotation has a contravariant generic parameter or a `super` constraint --
 * now the definitive type is the subtype.
 *
 * I considered splitting this out into a separate function and swapping the
 * order of the super / sub types passed to it, so we would only have to handle
 * one set of cases, but it doesn't look much better since that function still
 * has to recursively call sub_type and therefore needs to remember whether its
 * arguments had been swapped.
 *)
(****************************************************************************)
  | (r_sub, _), (_, Tunresolved _) ->
      let ty_sub = (r_sub, Tunresolved [ty_sub]) in
      sub_type_inner env ~this_ty ty_sub ty_super
   (* This case is for when Tany comes from expanding an empty Tvar - it will
    * result in binding the type variable to the other type. *)
  | _, (_, Tany) ->
    fst (Unify.unify env ty_super ty_sub)
    (* If the subtype is a type variable bound to an empty unresolved, record
     * this in the todo list to be checked later. But make sure that we wrap
     * any error using the position and reason information that was supplied
     * at the entry point to subtyping.
     *)
  | (_, Tunresolved []), _ ->
    begin match ty_sub with
    | (_, Tvar _) ->
      if not
        (TypecheckerOptions.experimental_feature_enabled
        (Env.get_tcopt env)
        TypecheckerOptions.experimental_unresolved_fix)
      then env
      else
        let outer_pos = env.Env.outer_pos in
        let outer_reason = env.Env.outer_reason in
        log_subtype ~this_ty:None "add_todo" env ty_sub ty_super;
        Env.add_todo env begin fun env' ->
          Errors.try_add_err outer_pos (Reason.string_of_ureason outer_reason)
          (fun () ->
            sub_type env' ty_sub ty_super, false)
          (fun _ ->
            env', true (* Remove from todo list if there was an error *)
          )
          end

    | _ ->
      env
    end

  | (_, Tunresolved tyl), _ ->
    let env =
      List.fold_left tyl ~f:begin fun env x ->
        sub_type_inner env ~this_ty
          x ty_super
      end ~init:env in
    env

(****************************************************************************)
(* ### End Tunresolved madness ### *)
(****************************************************************************)
  | (_, Tany), _ -> env

  (* This is sort of a hack because our handling of Toption is highly
   * dependent on how the type is structured. When we see a bare
   * dependent type we strip it off at this point since it shouldn't be
   * relevant to subtyping any more.
   *)
  | (_, Tabstract (AKdependent (`expr _, []), Some ty_sub)), _ ->
      let this_ty = Option.first_some this_ty (Some ety_sub) in
      sub_type_inner env ~this_ty ty_sub ty_super

  | (_, Tabstract (AKdependent d_sub, Some sub)),
    (_, Tabstract (AKdependent d_super, _)) when d_sub <> d_super ->
      let this_ty = Option.first_some this_ty (Some ety_sub) in
      (* If an error occurred while subtyping, we produce a unification error
       * so we get the full information on how the dependent type was
       * generated
       *)
      Errors.try_when
        (fun () ->
          sub_type_inner env ~this_ty
            sub ty_super)
        ~when_: begin fun () ->
          match sub, TUtils.get_base_type env ty_super with
          | (_, Tclass ((_, y), _)), (_, Tclass ((_, x), _)) when x = y -> false
          | _, _ -> true
        end
        ~do_: (fun _ -> TUtils.simplified_uerror env ty_super ty_sub)

  | (r_sub, _),
    (_, Tabstract (AKdependent (expr_dep, _),
      Some (_, Tclass ((_, x) as id, _) as ty_bound))) ->
    let class_ = Env.get_class env x in
    (* For final non-contravariant class C, there is no difference between
     * `this as C<t1,...,tn>` and `C<t1,...,tn>`.
     *)
    begin match class_ with
    | Some class_ty
      when TUtils.class_is_final_and_not_contravariant class_ty ->
      sub_type_inner env ~this_ty ty_sub ty_bound
    | _ ->
      (Errors.try_when
         (fun () -> TUtils.simplified_uerror env ty_super ty_sub)
             ~when_: begin fun () ->
               match snd ty_sub with
               | Tclass ((_, y), _) -> y = x
               | Tany | Terr | Tnonnull | Tarraykind _ | Tprim _
               | Toption _ | Tvar _ | Tabstract (_, _) | Ttuple _
               | Tanon (_, _) | Tfun _ | Tunresolved _ | Tobject
               | Tshape _ | Tdynamic -> false
             end
             ~do_: begin fun error ->
               if expr_dep = `cls x then
                 Errors.exact_class_final id (Reason.to_pos r_sub) error
               else
                 Errors.this_final id (Reason.to_pos r_sub) error
             end
          );
          env
    end

  | (p_sub, (Tclass (x_sub, tyl_sub))),
    (_, (Tclass (x_super, _tyl_super))) ->
    let cid_super, cid_sub = (snd x_super), (snd x_sub) in
    if cid_super = cid_sub
    (* Already dealt with this case (variance) in simplify_subtype *)
    then fst (Unify.unify env ety_super ety_sub)
    else
      let class_ = Env.get_class env cid_sub in
      begin match class_ with
        | None -> env
        | Some class_ ->
          if class_.tc_kind = Ast.Ctrait || class_.tc_kind = Ast.Cinterface then
          (* NOTE: We rely on the fact that we fold all ancestors of
           * ty_sub in its class_type so we will never hit this case
           * again. If this ever changes then we would need to store
           * ty_sub as the 'this_ty' in the uenv and be careful to
           * thread it through.
           *
           * This is covered by test/typecheck/this_tparam2.php
          *)
          let ety_env =
          {
            type_expansions = [];
            substs = Subst.make class_.tc_tparams tyl_sub;
            this_ty = Option.value this_ty ~default:ty_sub;
            from_class = None;
            validate_dty = None;
          } in
          let rec try_reqs reqs =
              match reqs with
              | [] ->
                fail ()

              | (_, req_type) :: reqs ->

              (* a trait is never the runtime type, but it can be used
               * as a constraint if it has requirements for its using
               * classes *)
                Errors.try_ begin fun () ->
                  let env, req_type =
                    Phase.localize ~ety_env env req_type in
                  sub_type env (p_sub, snd req_type) ty_super
                end (fun _ -> try_reqs reqs)
            in
              try_reqs class_.tc_req_ancestors
          else fail ()
    end

  | (_, Tdynamic), (_, Toption ty_super) ->
    sub_type_inner env ~this_ty ty_sub ty_super

  | (_, Tabstract (AKdependent _, Some ty)), (_, Toption arg_ty_super) ->
    Errors.try_
      (fun () ->
        sub_type_inner env ~this_ty ty_sub arg_ty_super)
      (fun _ ->
        let this_ty = Option.first_some this_ty (Some ety_sub) in
         sub_type_inner env ~this_ty ty ty_super)

  | (_, Tabstract (AKdependent _, Some ty)), _ ->
    let this_ty = Option.first_some this_ty (Some ety_sub) in
    sub_type_inner env ~this_ty ty ty_super

  | _, _ ->
    (* TODO: replace this by fail() once we support all subtyping rules that
     * are implemented in unification *)
     fst (Unify.unify env ty_super ty_sub)

(* BEWARE: hack upon hack here.
 * To implement a predicate that tests whether `ty_sub` is a subtype of
 * `ty_super`, we call sub_type but handle any unification errors and
 * turn them into `false` result. Unfortunately HH_FIXME might end up
 * hiding the "error", and so we need to disable the fixme mechanism
 * before calling sub_type and then re-enable it afterwards.
 *)
and is_sub_type
  (env : Env.env)
  (ty_sub : locl ty)
  (ty_super : locl ty) : bool =
  (* quick short circuit to help perf *)
  ty_equal ty_sub ty_super ||
  begin
    let f = !Errors.is_hh_fixme in
    Errors.is_hh_fixme := (fun _ _ -> false);
    let result =
      Errors.try_
        (fun () -> ignore(sub_type env ty_sub ty_super); true)
        (fun _ -> false) in
    Errors.is_hh_fixme := f;
    result
  end

and is_sub_type_alt env ty1 ty2 =
  let _env, prop =
    simplify_subtype ~seen_generic_params:None ~deep:true
    ~this_ty:(Some ty1) ty1 ty2 env in
  if TL.is_valid prop then Some true
  else
  if TL.is_unsat prop then Some false
  else None

(* Attempt to compute the intersection of a type with an existing list intersection.
 * If try_intersect env t [t1;...;tn] = [u1; ...; um]
 * then u1&...&um must be the greatest lower bound of t and t1&...&tn wrt subtyping.
 * For example:
 *   try_intersect nonnull [?C] = [C]
 *   try_intersect t1 [t2] = [t1]  if t1 <: t2
 * Note: it's acceptable to return [t;t1;...;tn] but the intention is that
 * we simplify (as above) wherever practical.
 * It can be assumed that the original list contains no redundancy.
 *)
and try_intersect env ty tyl =
  match tyl with
  | [] -> [ty]
  | ty'::tyl' ->
    if is_sub_type_alt env ty ty' = Some true
    then try_intersect env ty tyl'
    else
    if is_sub_type_alt env ty' ty = Some true
    then try_intersect env ty' tyl'
    else
    let nonnull_ty = (fst ty, Tnonnull) in
    match ty, ty' with
    | (_, Toption t), _ when is_sub_type_alt env ty' nonnull_ty = Some true ->
      try_intersect env t (ty'::tyl')
    | _, (_, Toption t) when is_sub_type_alt env ty nonnull_ty = Some true ->
      try_intersect env t (ty::tyl')
    | _, _ -> ty' :: try_intersect env ty tyl'

(* Attempt to compute the union of a type with an existing list union.
 * If try_union env t [t1;...;tn] = [u1;...;um]
 * then u1|...|um must be the least upper bound of t and t1|...|tn wrt subtyping.
 * For example:
 *   try_union int [float] = [num]
 *   try_union t1 [t2] = [t1] if t2 <: t1
 *
 * Notes:
 * 1. It's acceptable to return [t;t1;...;tn] but the intention is that
 *    we simplify (as above) wherever practical.
 * 2. Do not use Tunresolved for a syntactic union - the caller can do that.
 * 3. It can be assumed that the original list contains no redundancy.
 * TODO: there are many more unions to implement yet.
 *)
let rec try_union env ty tyl =
  match tyl with
  | [] -> [ty]
  | ty'::tyl' ->
    if is_sub_type_alt env ty ty' = Some true
    then try_union env ty' tyl'
    else
    if is_sub_type_alt env ty' ty = Some true
    then try_union env ty tyl'
    else match snd ty, snd ty' with
    | Tprim Nast.Tfloat, Tprim Nast.Tint
    | Tprim Nast.Tint, Tprim Nast.Tfloat ->
      let t = (fst ty, Tprim Nast.Tnum) in
      try_union env t tyl'
    | _, _ -> ty' :: try_union env ty tyl'

let rec sub_string
  ?(allow_mixed = false)
  (p : Pos.Map.key)
  (env : Env.env)
  (ty2 : locl ty) : Env.env =
  let stringish_deprecated =
    TypecheckerOptions.disallow_stringish_magic (Env.get_tcopt env) in
  let sub_string = sub_string ~allow_mixed in
  let env, ety2 = Env.expand_type env ty2 in
  let fail () =
    TUtils.uerror (Reason.Rwitness p) (Tprim Nast.Tstring) (fst ety2) (snd ety2);
    env in

  match ety2 with
  | (_, Toption ty2) -> sub_string p env ty2
  | (_, Tunresolved tyl) ->
      List.fold_left tyl ~f:(sub_string p) ~init:env
  | (_, Tprim _) ->
      env
  | (_, Tabstract (AKenum _, _)) ->
      (* Enums are either ints or strings, and so can always be used in a
       * stringish context *)
      env
  | (_, Tabstract _) ->
    begin match TUtils.get_concrete_supertypes env ty2 with
      | _, [] when allow_mixed ->
        env
      | _, [] ->
        fail ()
      | env, tyl ->
        List.fold_left tyl ~f:(sub_string p) ~init:env
    end
  | (r2, Tclass (x, _)) ->
      let class_ = Env.get_class env (snd x) in
      (match class_ with
      | None -> env
      | Some tc
          (* A Stringish is a string or an object with a __toString method
           * that will be converted to a string *)
          when tc.tc_name = SN.Classes.cStringish
          || SMap.mem SN.Classes.cStringish tc.tc_ancestors ->
        if stringish_deprecated
        then Errors.object_string_deprecated p;
        env
      | Some _ ->
        Errors.object_string p (Reason.to_pos r2);
        env
      )
  | _, (Tany | Terr | Tdynamic) ->
    env (* Tany, Terr and Tdynamic are considered Stringish by default *)
  | _, Tobject -> env
  | _, Tnonnull when allow_mixed ->
    env
  | _, (Tnonnull | Tarraykind _ | Tvar _
    | Ttuple _ | Tanon (_, _) | Tfun _ | Tshape _) ->
  fail ()

(** Check that the method with signature ft_sub can be used to override
 * (is a subtype of) method with signature ft_super.
 *
 * This goes beyond subtyping on function types because methods can have
 * generic parameters with bounds, and `where` constraints.
 *
 * Suppose ft_super is of the form
 *    <T1 csuper1, ..., Tn csupern>(tsuper1, ..., tsuperm) : tsuper where wsuper
 * and ft_sub is of the form
 *    <T1 csub1, ..., Tn csubn>(tsub1, ..., tsubm) : tsub where wsub
 * where csuperX and csubX are constraints on type parameters and wsuper and
 * wsub are 'where' constraints. Note that all types in the superclass,
 * including constraints (so csuperX, tsuperX, tsuper and wsuper) have had
 * any class type parameters instantiated appropriately according to
 * the actual arguments of the superclass. For example, suppose we have
 *
 *   class Super<T> {
 *     function foo<Tu as A<T>>(T $x) : B<T> where T super C<T>
 *   }
 *   class Sub extends Super<D> {
 *     ...override of foo...
 *   }
 * then the actual signature in the superclass that we need to check is
 *     function foo<Tu as A<D>>(D $x) : B<D> where D super C<D>
 * Note in particular the general form of the 'where' constraint.
 *
 * (Currently, this instantiation happens in
 *   Typing_extends.check_class_implements which in turn calls
 *   Decl_instantiate.instantiate_ce)
 *
 * Then for ft_sub to be a subtype of ft_super it must be the case that
 * (1) tsuper1 <: tsub1, ..., tsupern <: tsubn (under constraints
 *     T1 csuper1, ..., Tn csupern and wsuper).
 *
 *     This is contravariant subtyping on parameter types.
 *
 * (2) tsub <: tsuper (under constraints T1 csuper1, ..., Tn csupern and wsuper)
 *     This is covariant subtyping on result type. For constraints consider
 *       e.g. consider ft_super = <T super I>(): T
 *                 and ft_sub = <T>(): I
 *
 * (3) The constraints for ft_super entail the constraints for ft_sub, because
 *     we might be calling the function having checked that csuperX are
 *     satisfied but the definition of the function (e.g. consider an override)
 *     has been checked under csubX.
 *     More precisely, we must assume constraints T1 csuper1, ..., Tn csupern
 *     and wsuper, and check that T1 satisfies csub1, ..., Tn satisfies csubn
 *     and that wsub holds under those assumptions.
 *)
let subtype_method
  ~(check_return : bool)
  ~(extra_info: reactivity_extra_info)
  (env : Env.env)
  (r_sub : Reason.t)
  (ft_sub : decl fun_type)
  (r_super : Reason.t)
  (ft_super : decl fun_type) : Env.env =
  if not ft_super.ft_abstract && ft_sub.ft_abstract then
    (* It is valid for abstract class to extend a concrete class, but it cannot
     * redefine already concrete members as abstract.
     * See override_abstract_concrete.php test case for example. *)
    Errors.abstract_concrete_override ft_sub.ft_pos ft_super.ft_pos `method_;
  let ety_env =
    Phase.env_with_self env in
  let env, ft_super_no_tvars, _ =
    Phase.localize_ft ~use_pos:ft_super.ft_pos ~ety_env ~instantiate_tparams:false env ft_super in
  let env, ft_sub_no_tvars, _ =
    Phase.localize_ft ~use_pos:ft_sub.ft_pos ~ety_env ~instantiate_tparams:false env ft_sub in
  let old_tpenv = env.Env.lenv.Env.tpenv in

  (* We check constraint entailment and contravariant parameter/covariant result
   * subtyping in the context of the ft_super constraints. But we'd better
   * restore tpenv afterwards *)
  let add_tparams_constraints env (tparams: locl tparam list) =
    let add_bound env (_, (pos, name), cstrl, _) =
      List.fold_left cstrl ~init:env ~f:(fun env (ck, ty) ->
        let tparam_ty = (Reason.Rwitness pos,
          Tabstract(AKgeneric name, None)) in
        Typing_utils.add_constraint pos env ck tparam_ty ty) in
    List.fold_left tparams ~f:add_bound ~init: env in

  let p_sub = Reason.to_pos r_sub in

  let add_where_constraints env (cstrl: locl where_constraint list) =
    List.fold_left cstrl ~init:env ~f:(fun env (ty1, ck, ty2) ->
      Typing_utils.add_constraint p_sub env ck ty1 ty2) in

  let env =
    add_tparams_constraints env ft_super_no_tvars.ft_tparams in
  let env =
    add_where_constraints env ft_super_no_tvars.ft_where_constraints in

  let env, res =
  simplify_subtype_funs
    ~seen_generic_params:empty_seen
    ~deep:true
    ~check_return
    ~extra_info
    r_sub ft_sub_no_tvars
    r_super ft_super_no_tvars
    env in
  let env =
    process_simplify_subtype_result ~this_ty:None
    ~fail:(fun () -> ()) env res in

  (* This is (3) above *)
  let check_tparams_constraints env tparams =
  let check_tparam_constraints env (_var, (p, name), cstrl, _) =
    List.fold_left cstrl ~init:env ~f:begin fun env (ck, cstr_ty) ->
      let tgeneric = (Reason.Rwitness p, Tabstract (AKgeneric name, None)) in
      Typing_generic_constraint.check_constraint env ck cstr_ty tgeneric
    end in
  List.fold_left tparams ~init:env ~f:check_tparam_constraints in

  let check_where_constraints env cstrl =
    List.fold_left cstrl ~init:env ~f:begin fun env (ty1, ck, ty2) ->
      Typing_generic_constraint.check_constraint env ck ty2 ty1
    end in

  (* We only do this if the ft_tparam lengths match. Currently we don't even
   * report this as an error, indeed different names for type parameters.
   * TODO: make it an error to override with wrong number of type parameters
  *)
  let env =
    if List.length ft_sub.ft_tparams <> List.length ft_super.ft_tparams
    then env
    else check_tparams_constraints env ft_sub_no_tvars.ft_tparams in
  let env =
    check_where_constraints env ft_sub_no_tvars.ft_where_constraints in

  Env.env_with_tpenv env old_tpenv

let decompose_subtype_add_bound
  (env : Env.env)
  (ty_sub : locl ty)
  (ty_super : locl ty)
  : Env.env =
  let env, ty_super = Env.expand_type env ty_super in
  let env, ty_sub = Env.expand_type env ty_sub in
  match ty_sub, ty_super with
  (* name_sub <: ty_super so add an upper bound on name_sub *)
  | (_, Tabstract (AKgeneric name_sub, _)), _ when not (phys_equal ty_sub ty_super) ->
    log_subtype ~this_ty:None "decompose_subtype_add_bound" env ty_sub ty_super;
    let tys = Env.get_upper_bounds env name_sub in
    (* Don't add the same type twice! *)
    if Typing_set.mem ty_super tys then env
    else Env.add_upper_bound ~intersect:(try_intersect env) env name_sub ty_super

  (* ty_sub <: name_super so add a lower bound on name_super *)
  | _, (_, Tabstract (AKgeneric name_super, _)) when not (phys_equal ty_sub ty_super) ->
    log_subtype ~this_ty:None "decompose_subtype_add_bound" env ty_sub ty_super;
    let tys = Env.get_lower_bounds env name_super in
    (* Don't add the same type twice! *)
    if Typing_set.mem ty_sub tys then env
    else Env.add_lower_bound ~union:(try_union env) env name_super ty_sub

  | _, _ ->
    env

(* Given two types that we know are in a subtype relationship
 *   ty_sub <: ty_super
 * add to env.tpenv any bounds on generic type parameters that must
 * hold for ty_sub <: ty_super to be valid.
 *
 * For example, suppose we know Cov<T> <: Cov<D> for a covariant class Cov.
 * Then it must be the case that T <: D so we add an upper bound D to the
 * bounds for T.
 *
 * Although some of this code is similar to that for sub_type_inner, its
 * purpose is different. sub_type_inner takes two types t and u and makes
 * updates to the substitution of type variables (through unification) to
 * make t <: u true.
 *
 * decompose_subtype takes two types t and u for which t <: u is *assumed* to
 * hold, and makes updates to bounds on generic parameters that *necessarily*
 * hold in order for t <: u.
 *)
let rec decompose_subtype
  p
  (env : Env.env)
  (ty_sub : locl ty)
  (ty_super : locl ty)
  : Env.env =
  log_subtype ~this_ty:None "decompose_subtype" env ty_sub ty_super;
  let env, prop =
    simplify_subtype ~seen_generic_params:None
      ~deep:true ~this_ty:None ty_sub ty_super env in
  decompose_subtype_add_prop p env prop

and decompose_subtype_add_prop p env prop =
  match prop with
  | TL.Conj props ->
    List.fold_left ~f:(decompose_subtype_add_prop p) ~init:env props
  | TL.Disj _props ->
    Typing_log.log_prop 2 env.Env.pos "decompose_subtype_add_prop" env prop;
    env
  | TL.Unsat _ ->
    env
  | TL.IsSubtype (ty1, ty2) ->
    decompose_subtype_add_bound env ty1 ty2
  | TL.IsEqual (ty1, ty2) ->
    let env = decompose_subtype_add_bound env ty1 ty2 in
    decompose_subtype_add_bound env ty2 ty1

(* Decompose a general constraint *)
and decompose_constraint
  p
  (env : Env.env)
  (ck : Ast.constraint_kind)
  (ty_sub : locl ty)
  (ty_super : locl ty)
  : Env.env =
  match ck with
  | Ast.Constraint_as ->
    decompose_subtype p env ty_sub ty_super
  | Ast.Constraint_super ->
    decompose_subtype p env ty_super ty_sub
  | Ast.Constraint_eq ->
    let env' = decompose_subtype p env ty_sub ty_super in
    decompose_subtype p env' ty_super ty_sub

(* Given a constraint ty1 ck ty2 where ck is AS, SUPER or =,
 * add bounds to type parameters in the environment that necessarily
 * must hold in order for ty1 ck ty2.
 *
 * First, we invoke decompose_constraint to add initial bounds to
 * the environment. Then we iterate, decomposing constraints that
 * arise through transitivity across bounds.
 *
 * For example, suppose that env already contains
 *   C<T1> <: T2
 * for some covariant class C. Now suppose we add the
 * constraint "T2 as C<T3>" i.e. we end up with
 *   C<T1> <: T2 <: C<T3>
 * Then by transitivity we know that T1 <: T3 so we add this to the
 * environment also.
 *
 * We repeat this process until no further bounds are added to the
 * environment, or some limit is reached. (It's possible to construct
 * types that expand forever under inheritance.)
*)
let constraint_iteration_limit = 20
let add_constraint
  p
  (env : Env.env)
  (ck : Ast.constraint_kind)
  (ty_sub : locl ty)
  (ty_super : locl ty): Env.env =
  log_subtype ~this_ty:None "add_constraint" env ty_sub ty_super;
  let oldsize = Env.get_tpenv_size env in
  let env' = decompose_constraint p env ck ty_sub ty_super in
  if Env.get_tpenv_size env' = oldsize
  then env'
  else
  let rec iter n env =
    if n > constraint_iteration_limit then env
    else
      let oldsize = Env.get_tpenv_size env in
      let env' =
        List.fold_left (Env.get_generic_parameters env) ~init:env
          ~f:(fun env x ->
            List.fold_left (Typing_set.elements (Env.get_lower_bounds env x)) ~init:env
              ~f:(fun env ty_sub' ->
                List.fold_left (Typing_set.elements (Env.get_upper_bounds env x)) ~init:env
                  ~f:(fun env ty_super' ->
                    decompose_subtype p env ty_sub' ty_super'))) in
      if Env.get_tpenv_size env' = oldsize
      then env'
      else iter (n+1) env'
  in
    iter 0 env'

let flip_variance v =
  match v with
  | Ast.Covariant -> Ast.Contravariant
  | Ast.Contravariant -> Ast.Covariant
  | Ast.Invariant -> Ast.Invariant

let combine_variance v1 v2 =
  match v1, v2 with
  | Ast.Contravariant, v | v, Ast.Contravariant -> flip_variance v
  | Ast.Invariant, _ | _, Ast.Invariant -> Ast.Invariant
  | _ -> Ast.Covariant

(* For type variables in vars, set `appears_covariantly` and
 * `appears_contravariantly` in the type variable environment according to their
 * position in `ty`. The current position of `ty` is indicated by `variance`.
 *)
let rec set_tyvar_variance ~variance ~tyvars env ty =
  match snd ty with
  | Tvar v ->
    if ISet.mem v tyvars
    then
      match variance with
      | Ast.Covariant -> Env.set_tyvar_appears_covariantly env v
      | Ast.Contravariant -> Env.set_tyvar_appears_contravariantly env v
      | Ast.Invariant ->
        let env = Env.set_tyvar_appears_covariantly env v in
        Env.set_tyvar_appears_contravariantly env v
    else env
  | Tany | Tnonnull | Terr | Tdynamic | Tobject | Tprim _ | Tanon _ | Tabstract(_, None) ->
    env
    (* Nullable is covariant *)
  | Toption ty ->
    set_tyvar_variance ~variance ~tyvars env ty
    (* Tuples and unions are covariant *)
  | Ttuple tyl | Tunresolved tyl ->
    List.fold_left ~f:(set_tyvar_variance ~variance ~tyvars) ~init:env tyl
    (* Shape data is covariant *)
  | Tshape (_, m) ->
    Nast.ShapeMap.fold begin fun _ { sft_ty; _ } env ->
      set_tyvar_variance ~variance ~tyvars env sft_ty end m env
    (* Functions are covariant in return type, contravariant in parameter types *)
  | Tfun ft ->
    let flipped = flip_variance variance in
    let env = List.fold_left ~f:begin fun env { fp_type; _ } ->
      set_tyvar_variance ~variance:flipped ~tyvars env fp_type end ~init:env ft.ft_params in
    set_tyvar_variance ~variance ~tyvars env ft.ft_ret
  | Tabstract (AKnewtype (name, tyl), _) ->
    begin match Env.get_typedef env name with
    | Some {td_tparams; _} ->
      let variancel = List.map td_tparams (fun (v,_,_,_) -> combine_variance variance v) in
      set_tyvar_variance_list ~variancel ~tyvars env tyl
    | None ->
      env
    end
  | Tabstract (_, Some ty) ->
    set_tyvar_variance ~variance ~tyvars env ty
    (* Classes carry their own variance declarations *)
  | Tclass ((_, cid), tyl) ->
    begin match Env.get_class env cid with
    | None -> env
    | Some {tc_tparams; _} ->
      let variancel = List.map tc_tparams (fun (v,_,_,_) -> combine_variance variance v) in
      set_tyvar_variance_list ~variancel ~tyvars env tyl
    end
    (* Arrays are covariant in key and data types *)
  | Tarraykind ak ->
    begin match ak with
    | AKany | AKempty -> env
    | AKvarray ty | AKvec ty | AKvarray_or_darray ty ->
      set_tyvar_variance ~variance ~tyvars env ty
    | AKdarray (ty1, ty2) | AKmap (ty1, ty2) ->
      let env = set_tyvar_variance ~variance ~tyvars env ty1 in
      set_tyvar_variance ~variance ~tyvars env ty2
    | AKshape m ->
      Nast.ShapeMap.fold begin fun _ (ty1, ty2) env ->
        let env = set_tyvar_variance ~variance ~tyvars env ty1 in
        set_tyvar_variance ~variance ~tyvars env ty2 end m env
    | AKtuple m ->
      IMap.fold (fun _ ty env -> set_tyvar_variance ~variance ~tyvars env ty) m env
    end

and set_tyvar_variance_list ~variancel ~tyvars env tyl =
  match variancel, tyl with
  | [], [] -> env
  | variance::variancel, ty::tyl ->
    let env = set_tyvar_variance ~variance ~tyvars env ty in
    set_tyvar_variance_list ~variancel ~tyvars env tyl
  | _ -> env

let set_tyvar_variance ~tyvars env ty =
  set_tyvar_variance ~variance:Ast.Covariant ~tyvars env ty

let remove_tyvar env lower_bounds var upper_bounds =
  let env =
    List.fold_left lower_bounds ~init:env
    ~f:(fun env ty_sub ->
      List.fold_left upper_bounds ~init:env
        ~f:(fun env ty_super ->
          sub_type env ty_sub ty_super)) in
  Env.remove_tyvar env var

(* Use the variance information about a type variable to force a solution.
 *   (1) If the type variable is bounded by t1, ..., tn <: v and it appears only
 *   covariantly or not at all in the expression type then
 *   we can minimize it i.e. set v := t1 | ... | tn.
 *   (2) If the type variable is bounded by v <: t1, ..., tn and it appears only
 *   contravariantly in the expression type then we can maximize it. Ideally we
 *   would use an intersection v := t1 & ... & tn so for now we only solve
 *   if there is a single upper bound.
 *)
let solve_tyvar env r var =
  let appears_contravariantly = Env.get_tyvar_appears_contravariantly env var in
  let appears_covariantly = Env.get_tyvar_appears_covariantly env var in
  (* As in Local Type Inference by Pierce & Turner, if type variable
   * appears only contravariantly, force to upper bound
   *)
  if appears_contravariantly && not appears_covariantly
  then begin
    let upper_bounds = Typing_set.elements (Env.get_tyvar_upper_bounds env var) in
    let lower_bounds = Typing_set.elements (Env.get_tyvar_lower_bounds env var) in
    match upper_bounds with
    | [ty] ->
      Typing_log.(log_with_level env "prop" 2 (fun () ->
      log_types (Reason.to_pos r) env
        [Log_head (Printf.sprintf "Typing_subtype.solve_tyvar/Typing_unify_recursive.add #%d" var,
        [Log_type ("ty", ty)])]));
      (* Unify the variable with the upper bound *)
      let env = Typing_unify_recursive.add env var ty in
      (* Remove the variable from the environment. This also has the effect
       * of closing transitively i.e. checking that lower bounds are sutypes of upper bounds.
       *)
      remove_tyvar env lower_bounds var upper_bounds
      (* For now, if there are no bounds, or more than one, then don't solve.
       * For no bounds, we could just use `mixed`
       *)
    | _ ->
      env
  end
  (* As in Local Type Inference by Pierce & Turner, if type variable does
   * not appear at all, or only appears covariantly, force to lower bound
   *)
  else if not appears_contravariantly
  then begin
    let upper_bounds = Typing_set.elements (Env.get_tyvar_upper_bounds env var) in
    let lower_bounds = Typing_set.elements (Env.get_tyvar_lower_bounds env var) in
    if List.is_empty lower_bounds
    then env
    else
    (* Construct the union of the lower bounds *)
    let ty =
      match lower_bounds with
      | [ty] -> ty
      | tys -> (r, Tunresolved tys) in
    Typing_log.(log_with_level env "prop" 2 (fun () ->
      log_types (Reason.to_pos r) env
        [Log_head (Printf.sprintf "Typing_subtype.solve_tyvar/Typing_unify_recursive.add #%d" var,
        [Log_type ("ty", ty)])]));
    let env = Typing_unify_recursive.add env var ty in
    remove_tyvar env lower_bounds var upper_bounds
  end
  else env

let solve_tyvars ~tyvars env =
  List.fold_left ~init:env ~f:(fun env tyvar -> solve_tyvar env Reason.Rnone tyvar)
    (ISet.elements tyvars)

(*****************************************************************************)
(* Exporting *)
(*****************************************************************************)

let () = Typing_utils.sub_type_ref := sub_type
let () = Typing_utils.add_constraint_ref := add_constraint
let () = Typing_utils.is_sub_type_ref := is_sub_type
let () = Typing_utils.is_sub_type_alt_ref := is_sub_type_alt
